/*
 * $Id$
 * Author: Karoly Bosa
 *
 * (C) 2008 by Research Institute for Symbolic Computation, Linz, Austria
 * All rights reserved. See COPYRIGHT in top-level directory.
 */

#include <taagimpl.h>
#include <xml_parser.h>

int _taag_initialized = TAAG_FALSE;
int _hw_initialized = TAAG_FALSE;


static enum _schema_types type;

static unsigned int nrOfGrps = 0;
static unsigned int depth = 0;

static unsigned int *nrOfProcs;
static unsigned int **listOfProcs;

static int *distances;

/***** Hardware Data Struct *****/
static unsigned int nrOfLANs = 0;

static double *latencies; //among the hosts
static char **hostNames;
static int *nrOfCPUs; //on the hosts

//nr of all host contained by last element
//nr of hosts on i. LAN = i. element - (i-1). element
//hosts of i. LAN: (i-1). element  to i. element -1
//
//Advantage of this representation that we do not have to store
//the rank of the hosts;
static int *hostsOnLAN;

static int *nrOfProcsOnHosts;
static int **procsOnHosts;

/********************************/


enum _schema_types _taag_getType() {
	return (type);
}

int _taag_getNumberOfGroups() {
	return(nrOfGrps);
}

int _taag_getDepth() {
	return (depth);
}

int _taag_nrOfLAN() {
	return(nrOfLANs);
}

int _taag_nrOfHost() {
	return(*(hostsOnLAN + nrOfLANs -1));
}

int _taag_nrOfCPUs() {
	int counter = 0;
	int nrHost = _taag_nrOfHost();

	for (int i = 0; i < nrHost; i++) {
		counter += *(nrOfCPUs+i);
	}
	return (counter);
}

/***********************************************************************
 *
 ***********************************************************************/
int _taag_getHost(int rank) {
	int host = -1;
	int nrProcs;
	int *processes;
	int found = TAAG_FALSE;
	int nrHost = _taag_nrOfHost();

	for (int i = 0; i < nrHost && !found; i++) {
		nrProcs = *(nrOfProcsOnHosts + i);
		processes = *(procsOnHosts + i);

		for (int j = 0; j < nrProcs && !found; j++) {
			if (rank == *(processes+j)) {
				host = i;
				found = TAAG_TRUE;
			} //if
		} //for
	} //for
	return (host);
}

/***********************************************************************
 *
 ***********************************************************************/
int _taag_getLAN(int host) {
	int lan = -1;
	int hosts;
	int found = TAAG_FALSE;
	int nrLAN = _taag_nrOfLAN();

	for (int i = 0; i < nrLAN && !found; i++) {
		hosts = *(hostsOnLAN + i);
		if (host < hosts) {
			lan = i;
			found = TAAG_TRUE;
		} //if
	} //for

	return (lan);
}

/***********************************************************************
 *
 ***********************************************************************/
void *_taag_getPointer1(enum _pointer_types ptype, int grpRank) {
	if (grpRank < 0 || grpRank >= nrOfGrps) return NULL;
	switch(ptype) {
		case NR_PROCS: return(nrOfProcs + grpRank); break;
		case LIST_PROCS: return(listOfProcs + grpRank); break;
		case DISTANCES : return(distances + grpRank * nrOfGrps); break;

		default: return NULL;

	} //switch

}

/***********************************************************************
 *
 ***********************************************************************/
void *_taag_getPointer2(enum _pointer_types ptype, int host) {
	if (host < 0 || host >= _taag_nrOfHost()) return NULL;
	switch(ptype) {
		case LATENCIES : return(latencies + host * _taag_nrOfHost() ); break;
		case HOSTNAMES : return(hostNames + host); break;
		case CPUS      : return(nrOfCPUs + host); break;
		case NR_PROCS_ON_HOSTS : return(nrOfProcsOnHosts + host); break;
		case PROCS_ON_HOSTS : return(procsOnHosts + host); break;
		default: return NULL;

	} //switch

}

/***********************************************************************
 *
 ***********************************************************************/
void *_taag_getPointer3(enum _pointer_types ptype, int lan) {
	if (lan < 0 || lan >= nrOfLANs) return NULL;
	switch(ptype) {
		case HOSTS_ON_LANS : return(hostsOnLAN + lan); break;
		default: return NULL;

	} //switch

}

/***********************************************************************
 * This function computes the distance matrix from the neighbourhood
 * matrix of the graph
 ***********************************************************************/
static void createDistanceMatrix() {
	int found = TAAG_TRUE;

	for (int k = 1; k <nrOfGrps-1 && found; k++) { // k is the distance
		found = TAAG_FALSE;
		for (int i = 0; i < nrOfGrps; i++) { // i is the actual node

			for (int j = 0 ; j < nrOfGrps; j++) { //j denotes the k-th neighbors of i
				if (*(distances+i*nrOfGrps+j) != k) continue;

				for (int l = i; l < nrOfGrps; l++) { //l denotes the direct neighbors of j
					if (*(distances+j*nrOfGrps+l) == 1) {
						if (		*(distances+i*nrOfGrps+l) == -1
										||
									*(distances+i*nrOfGrps+l) > k+1) {
							*(distances+i*nrOfGrps+l) = k+1;
							*(distances+l*nrOfGrps+i) = k+1;
							found = TAAG_TRUE;
						} //if
					} //if
				} //for
			} //for
		} //for
	} //for
}

/***********************************************************************
 *
 ***********************************************************************/
static int computeDepth() {
	int max;

	max = *(distances + TAAG_ROOT*nrOfGrps);
	for (int i = 1; i < nrOfGrps; i++) {
		if (max < *(distances + TAAG_ROOT*nrOfGrps + i)) {
			max = *(distances + TAAG_ROOT*nrOfGrps + i);
		} //if
	} //for
	return (max+1);
}

/***********************************************************************
 * TAAG_Init allocates and initializes the corresponding data structures
 * according to the generated execution plan ﬁle given in the argument.
 * If the argument is null then the latest available execution plan is
 * taken. If there is none, then the call returns an error code which
 * is different as TAAG SUCCESS. This function must be called in
 * every program, must be called before any other TAAG functions and
 * must be called only once in a program.
 ***********************************************************************/
int TAAG_Init(char *file) {
	int flag;
	char s_data[TAAG_MAX_HOSTNAME_STRING];
	int tmp;

	MPI_Initialized(&flag);
	if (!flag) return (TAAG_ERR_MPI);

	if (_taag_xml_openFile(file)) return(TAAG_ERR_XML);
	if (_taag_xml_validate()) return(TAAG_ERR_XML);

	/*****get the data of the graph*****/
	//get type
	_taag_xml_getSchemaType(s_data);
	if (!strcmp(s_data, "groups")) type = SCHEMA_GROUPS;
    else if (!strcmp(s_data, "graph")) type = SCHEMA_GRAPH;
	else if (!strcmp(s_data, "tree")) type = SCHEMA_TREE;
	else if (!strcmp(s_data, "ring")) type = SCHEMA_RING;
	else return (TAAG_ERR_SCHEMA);

	//***** get number of groups
	_taag_xml_getNumberOfGroups(&nrOfGrps);

	//***** get sizes of all groups
	nrOfProcs = (unsigned int *)calloc(nrOfGrps, sizeof(unsigned int));
	_taag_xml_getGroupSizes(nrOfGrps, nrOfProcs);

	//***** get members of all groups
	listOfProcs = (unsigned int **)calloc(nrOfGrps, sizeof(unsigned int *));
	for (int i = 0; i < nrOfGrps; i++) {
		*(listOfProcs+i) = (unsigned int *)calloc( *(nrOfProcs + i), sizeof(unsigned int));
	} //for
	_taag_xml_getGroupMembers(nrOfGrps, nrOfProcs, listOfProcs);

	//***** get the neighborhood matrix of the whole graph
	tmp = nrOfGrps*nrOfGrps;
	distances = (int *)calloc(tmp, sizeof(int));
	//Initializing
	for (int i = 0; i < tmp; i++) {
		if (i/nrOfGrps != i%nrOfGrps){
			*(distances+i) = -1;
		} //if
		else {
			*(distances+i) = 0;
		} //else
	} //for

	_taag_xml_getEdgeMatrix(nrOfGrps, distances);

	/*****get the data of the topology*****/
	_taag_xml_isTopologyGiven(&_hw_initialized);
	if (_hw_initialized == TAAG_TRUE) {
		if (_taag_xml_getNumberOfLANs(&nrOfLANs)) return(TAAG_ERR_XML);

		hostsOnLAN = (int *)calloc(nrOfLANs, sizeof(int));
		_taag_xml_getHosts(nrOfLANs, hostsOnLAN);

		hostNames = (char **)calloc(_taag_nrOfHost(), sizeof(char *));
		for (int i = 0; i < _taag_nrOfHost(); i++) {
			*(hostNames+i) = (char *)calloc(TAAG_MAX_HOSTNAME_STRING, sizeof(char));
		}
		_taag_xml_getHostnames(_taag_nrOfHost(), hostNames);

		nrOfCPUs =	(int *)calloc(_taag_nrOfHost(), sizeof(int));
		memset(nrOfCPUs, 0, _taag_nrOfHost()*sizeof(int));
		_taag_xml_getNumberOfCPUs(_taag_nrOfHost(), nrOfCPUs);

		tmp = _taag_nrOfHost() * _taag_nrOfHost();
		latencies = (double *)calloc(tmp, sizeof(double));
		//Initializing
		for (int i = 0; i < tmp; i++) {
			*(latencies+i) = -1.0;
		} //for
		_taag_xml_getLatency(_taag_nrOfHost(), latencies);

		nrOfProcsOnHosts = (int *)calloc(_taag_nrOfHost(), sizeof(int));
		memset(nrOfProcsOnHosts, 0, _taag_nrOfHost()*sizeof(int));
		_taag_xml_getNumberOfProcsOnHosts(_taag_nrOfHost(), nrOfProcsOnHosts);

		//***** get processes located on all hosts
		procsOnHosts = (int **)calloc(_taag_nrOfHost(), sizeof(int *));
		for (int i = 0; i < _taag_nrOfHost(); i++) {
			*(procsOnHosts+i) = (int *)calloc( *(nrOfProcsOnHosts + i), sizeof(int));
			memset(*(procsOnHosts+i), 0, *(nrOfProcsOnHosts+i) * sizeof(int));
		} //for
		_taag_xml_getProcsOnHosts(_taag_nrOfHost(), nrOfProcsOnHosts, procsOnHosts);

	} //if

	//free xml parser
	_taag_xml_free();

	createDistanceMatrix();
	depth = computeDepth();

	_taag_initialized = TAAG_TRUE;
	return (TAAG_SUCCESS);
}

/***********************************************************************
 * TAAG_Initialized indicates whether TAAG Init has been called.
 * It returns a ﬂag as either logical true (1) or false(0).
 ***********************************************************************/
int TAAG_Initialized(int *flag) {
	*flag = _taag_initialized;
	return (TAAG_SUCCESS);
}

/***********************************************************************
 * TAAG_Free deallocates the data structures used by the API library.
 *
 ***********************************************************************/
int TAAG_Free() {
	unsigned int *tmp;
	int *tmp2;
	char *tmp3;

	if (_taag_initialized == TAAG_FALSE) return (TAAG_ERR_INIT);

	_taag_initialized = TAAG_FALSE;

	free(nrOfProcs);

	for (int i = 0; i < nrOfGrps; i++) {
		tmp = *(listOfProcs+i);
		free(tmp);
	} //for
	free(listOfProcs);

    free(distances);
    /***** free the topology related datas *****/
    if (_hw_initialized == TAAG_TRUE) {
		free(latencies);
		for (int i = 0; i < _taag_nrOfHost(); i++) {
			tmp3 = *(hostNames+i);
			free(tmp3);
		} //for
		free(hostNames);
		free(nrOfCPUs);
		free(nrOfProcsOnHosts);
		for (int i = 0; i < _taag_nrOfHost(); i++) {
			tmp2 = *(procsOnHosts+i);
			free(tmp2);
		} //for
		free(procsOnHosts);
		free(hostsOnLAN);
    }
    return (TAAG_SUCCESS);
}

