/*
 * $Id$
 * Author: Karoly Bosa
 *
 * (C) 2010 by Research Institute for Symbolic Computation, Linz, Austria
 * All rights reserved. See COPYRIGHT in top-level directory.
 */

#include "LatencyCluster.h"

LatencyCluster::LatencyCluster() {
	level = 0;
	contained = false;
	chosen = false;
	alreadyChosen = false;

}

LatencyCluster::LatencyCluster(int _level) {
	level = _level;
	contained = false;
	chosen = false;
	alreadyChosen = false;
}

LatencyCluster::~LatencyCluster() {

}

bool LatencyCluster::isEmpty() {
	return hostList.empty();
}

int LatencyCluster::addHost(HostData *newHost) {
	hostList.push_back(newHost);
	newHost->setAddressOfHighestCluster(this);
	if (level==0 || level==1) newHost->setAddressOfLANCluster(this);
	return hostList.size();
}

int LatencyCluster::getNrOfHosts() {
	return hostList.size();
}

double LatencyCluster::getNrOfCPUs() {
	if (hostList.empty()) return 0;
	double sumOfCPUFraction = 0.0;
	vector<HostData *>::iterator i;
	for (i = hostList.begin(); i != hostList.end(); i++) {
		sumOfCPUFraction += (*i)->getCPUForecast();
	}
	return sumOfCPUFraction;
}//	return hostList.size();

int LatencyCluster::getProcsCapacity() {
	int sumOfProcsSlot = 0;
	vector<HostData *>::iterator i;
	for (i = hostList.begin(); i != hostList.end(); i++) {
		sumOfProcsSlot += (*i)->availableProcsCapacity();
	}
	return sumOfProcsSlot;
}

bool LatencyCluster::areEnoughProcsCapacity(int nrOfProcs) {
	if (hostList.empty() || nrOfProcs<=0) return false;
	if (getProcsCapacity() >= nrOfProcs) return true;
	else return false;

}

bool LatencyCluster::mapGroupToHosts(GroupData *grp) {
	int nrOfProcs = grp->getNrOfProcs();
	if (!areEnoughProcsCapacity(nrOfProcs) || nrOfProcs<=0) return false;
	//Schedule given procs to hosts which have the lowest capacity!!!
	//host with greater capacity might fit to smaller group on level 0!!!!
	grp->setTmpLevel(level);
	do {
		vector<HostData *>::iterator it = hostList.begin();
		HostData *host = *it;
		it++;
		for (; it !=hostList.end(); it++ ) {
			if ((*it)->availableProcsCapacity() > 0
			   &&
			  (host->availableProcsCapacity() > (*it)->availableProcsCapacity()
			       ||
			   host->availableProcsCapacity()==0)) {
				host = *it;
			}
		}
		if (host->availableProcsCapacity()==0) {
			printf("Error: Inconsistent latency cluster.\n");
			exit(0);
		}
		if (nrOfProcs > host->availableProcsCapacity()) {
			nrOfProcs -= host->availableProcsCapacity();
			grp->createLoadTicket(host->availableProcsCapacity(), host);
		}
		else {
			grp->createLoadTicket(nrOfProcs, host);
			nrOfProcs = 0;
		}
	} while(nrOfProcs > 0);
	return true;
}

/************************************
 * However the number of tree managers can be greater
 * than some of the leaf groups, they can be mapped
 * later than the leaf groups (because leaf groups have the
 * priority for optimal mapping)
 ************************************/
bool LatencyCluster::mapMgrsToHosts(ProcsDistribution *pd, int firstMgrIndex) {
	if (!areEnoughProcsCapacity(firstMgrIndex+1)) return false;
	for (int grpIndex = firstMgrIndex; grpIndex >= 0; grpIndex--) {
		GroupData *grp = pd->getGroupByIndex(grpIndex);
		grp->setTmpLevel(level);

		//Schedule given procs to hosts which have the lowest capacity!!!
		//host with greater capacity might fit to smaller group on level 0!!!!
		vector<HostData *>::iterator it = hostList.begin();
		HostData *host = *it;
		it++;
		for (; it !=hostList.end(); it++ ) {
			if ((*it)->availableProcsCapacity() > 0
			   &&
			  (host->availableProcsCapacity() > (*it)->availableProcsCapacity()
				   ||
			   host->availableProcsCapacity()==0)) {
				host = *it;
			}
		}
		if (host->availableProcsCapacity()==0) {
			printf("Error: Inconsistent latency cluster.\n");
			exit(0);
		}
		grp->createLoadTicket(1, host); //every manager group contains always onlt one process
	}
	return true;
}

void LatencyCluster::setLevel(int _level) {
	level = _level;
}

int LatencyCluster::getLevel() {
	return level;
}

HostData *LatencyCluster::getHostByIndex(unsigned int index) {
	if (hostList.empty() || index >= hostList.size() || index < 0) return NULL;
	return hostList[index];
}

void LatencyCluster::printOutLatencyCluster(FILE *f) {
	if (hostList.empty()) return;
		fprintf(f, "%d. level (%5.2f CPUs)\t[", level, getNrOfCPUs());
		vector<HostData *>::iterator i;
		for (i = hostList.begin(); i != hostList.end(); i++) {
			fprintf(f, " %s", (*i)->getHostname());
	}
	fprintf(f, " ]\n");
}

void LatencyCluster::setContained() {
	contained =  true;
}

bool LatencyCluster::isContained() {
	return contained;
}

void LatencyCluster::setChosen(bool value) {
	chosen = value;
	alreadyChosen = true;
	//printOutLatencyCluster(stdout); //BK_TEST
	vector<LatencyCluster *>::iterator it;
	for (it = clusterList.begin(); it != clusterList.end(); it++) {
		(*it)->setChosen(value);
	}
}

bool LatencyCluster::isChosen() {
	return chosen;
}

bool LatencyCluster::wasChosen() {
	return alreadyChosen;
}

int LatencyCluster::addCluster(LatencyCluster *newCluster) {
	clusterList.push_back(newCluster);
	newCluster->setContained();
	return clusterList.size();
}

//int LatencyCluster::getNtOfCluster() {}
//LatencyCluster *LatencyCluster::getClusterByIndex(unsigned int index) {}

/***********************************************************
 * The latency Clusters are disjoint, so we have to
 * check only the first two hosts which belongs to the establishing
 * connection
 ***********************************************************/
void LatencyCluster::collectHostsFromContainedClusters() {
	vector<LatencyCluster *>::iterator it;
	int id1 = hostList[0]->getHostId();
	int id2 = hostList[1]->getHostId();
	for (it = clusterList.begin(); it != clusterList.end(); it++) {
		for (int i = 0; i < (*it)->getNrOfHosts(); i++) {
			int newId =(*it)->getHostByIndex(i)->getHostId();
			if (id1 != newId && id2 != newId) {
				addHost((*it)->getHostByIndex(i));
			}
		} //for i
	} //for it
}


